# TDAbletonExt
# 1.23.7

TDJ = op.TDModules.mod.TDJSON
TDF = op.TDModules.mod.TDFunctions
from datetime import datetime
from copy import deepcopy
from pathlib import Path
from distutils import version
import os
import os.path
import shutil
import subprocess
from xml.etree import ElementTree
import pprint

LOMPARS = mod.TDAbletonCompBaseExt.LOMPARS

class TDAbletonExt:

	"""
	TDAbletonExt provides the master connection to an Ableton Live process.
	"""

	# region Main

	def __init__(self, ownerComp):
		# The component to which this extension is attached
		self.ownerComp = ownerComp
		self.SourceComp = self.ownerComp.op('comps')
		self.Utils = self.ownerComp.op('Utils').module
		self.OSCIn = self.ownerComp.op('oscin')
		self.OSCOut = self.ownerComp.op('oscout')
		self.oscInTime = self.ownerComp.op('oscin_time')
		self.OSCInTime = self.ownerComp.op('time')
		self.OSCMaxOut = self.ownerComp.op('osc_maxOut')
		self.Console = self.ownerComp.op('console')
		self.TDABroadcast = self.ownerComp.op('tdaMaster/tdaBroadcast')
		self.TDABroadcastPort = self.ownerComp.op('tdaMaster/tdaBroadcastPort')
		self.TDAAbletonSendTime = self.ownerComp.op(
											'tdaMaster/tdaAbletonSendTimeData')
		self.connectingDialog = self.ownerComp.op('connectingDialog')

		# see songInfoStructure.py for layout (keyed by ptr STRINGS!)
		self.SongInfoByPtr = dict({'tracks':{}})
		# see songInfoStructure.py for layout (keyed by name)
		self.SongInfo = dict({'tracks':{}})
		# {<ptr>: {'byName': <objInfo in SongInfoByName>,
		# 			'byPtr': <objInfo in SongInfoByPtr>}...}
		self.ObjectInfoByPtr = {}

		# TDAbleton is connected...
		TDF.createProperty(self, 'Connected', value=False, readOnly=True,
															dependable=True)
		self.loadingSongInfo = '' # Used as song is loading
		self.InPort = None # port number
		self.retryConnectRepeater = None # to store run object
		self.listenOnlyPar = self.ownerComp.par.Listenonly # for speed
		self.remoteRequestDict = {} # {<requestRemoteData ID>: callback}
		self.defaultRemoteDataID = 0 # used as starting number
		# used to restore install if bad one is selected
		self.ownerComp.par.Updateabletoncomps.label = "Update All To " + \
				self.ownerComp.par.Version.eval()
		if hasattr(self.ownerComp.parent().par, 'Updateabletoncomps'):
			self.ownerComp.parent().par.Updateabletoncomps.label = \
					"Update All To " + \
								self.ownerComp.par.Version.eval()
		run('op(' + str(ownerComp.id) + ').ext.TDAbletonExt.setupConnection()'\
			'if op(' + str(ownerComp.id) + ') else None',
					delayFrames=1, delayRef=op.TDResources)

	def SendOSC(self, *args):
		# send blocker/tester
		if not self.Connected and not args[0].startswith('/tda/command'):
			return
		if self.listenOnlyPar:
			if not(args[0].startswith('/listener/add') or
						args[0].startswith('/listener/remove') or
							args[0].startswith('/tda/command') or
								args[0].startswith('/shell')):
				return
		self.OSCOut.sendOSC(*args)

	def SendMaxOSC(self, *args):
		if not self.Connected or self.listenOnlyPar:
			return
		self.OSCMaxOut.sendOSC(*args)

	def disconnect(self):
		self.oscInTime.par.resetchannels.pulse()
		if self._Connected:
			errorMsg = 'Disconnected from Ableton Live. (Press Help Page ' \
						'parameter for documentation)'
			self._Connected.val = False
			if self.ownerComp.par.Nodeerrorondisconnect \
								and self.ownerComp.par.Connect:
				self.SetError(errorMsg)
			self.SendOSC('/tda/command', ['disconnect', self.InPort])
		self.NotifyAbletonComps('disconnected')

	def connect(self, displayRequest=True):
		if self.Connected:
			return
		self.setDebugging()
		self.InPort = self.ownerComp.par.Inport.eval()
		self.SendOSC('/tda/command', ['connect', self.InPort,
										self.ownerComp.par.Version])
		if displayRequest:
			print(self.ownerComp.path,
					': Requesting connection to Ableton Live.')
			#self.connectingDialog.par.Open.pulse()
		try:
			self.retryConnectRepeater.kill()
		except:
			pass
		finally:
			self.retryConnectRepeater = None
		self.retryConnectRepeater = run(
					'op(' + str(self.ownerComp.id) + ').'
						'ext.TDAbletonExt.retryConnect(' +
								str(displayRequest) + ') '
							'if op(' + str(self.ownerComp.id) + ') else None',
				delayMilliSeconds=1000, delayRef=op.TDResources)

	def retryConnect(self, displayRequest=False):
		if displayRequest and self.ownerComp.par.Showconnectingdialog:
			self.connectingDialog.par.Open.pulse()
		if not self._Connected.val and self.ownerComp.par.Connect.eval():
			self.connect(False)

	def setupConnection(self):
		"""
		Set up for OSC communications with Ableton. Default is par.Inport
		And then reconnect to TDA.
		"""
		# if not self._Connected:
		# 	self.SetError("Can't connect to Ableton Live. (Press Help Page "
		# 			   		"parameter for documentation)")
		self.lastSelectedInstall = self.ownerComp.par.Abletonliveinstall.eval()
		self.disconnect()
		self.InPort = self.ownerComp.par.Inport.eval()
		self.OSCOut.par.address = self.ownerComp.par.Abletonaddress.eval()
		self.OSCOut.par.port = self.ownerComp.par.Abletonport.eval()
		self.OSCMaxOut.par.address = self.ownerComp.par.Abletonaddress.eval()
		self.OSCMaxOut.par.port = self.ownerComp.par.Maxport.eval()
		self.NotifyAbletonComps("reinit")
		if self.ownerComp.par.Connect.eval():
			run('op(' + str(self.ownerComp.id) + ').ext.TDAbletonExt.connect()',
				delayMilliSeconds=1000, delayRef=op.TDResources)

	def setDebugging(self):
		self.SendOSC('/tda/command', ['setDebugging',
						float(self.ownerComp.par.Logtdadebugmsgs.eval())])

	def SetError(self, error=''):
		if error == '':
			self.ownerComp.clearScriptErrors(error='TDAbleton:*')
		else:
			print(self.ownerComp.path + ":", error)
			self.ownerComp.addScriptError("TDAbleton: " + error)

	# endregion

	# region Ableton Comp Utilities

	@property
	def TDAbletonComps(self):
		return root.findChildren(tags=('TDAbleton',),
				key=lambda x: x.par.Tdabletoncomp.eval() == self.ownerComp or
									x.par.Tdabletoncomp.val == '')									

	@property
	def TDAMasterDevice(self):
		"""
		:return: The TDA_Master device on Ableton's master track
		"""
		for device in self.SongInfo['tracks']\
								['# Master Track #']['devices'].values():
			if 'ID_TDA_Master' in device['aPars']:
				return device
		return None

	@property
	def AbletonBroadcast(self):
		return self.ownerComp.par.Abletonbroadcast.eval()

	def UpdateAbletonComps(self, comps=None):
		"""
		Clone TDAbleton comps from their source in tdAbleton/comps.
		Do other minor duplication tasks

		:param comps: list of comps to be cloned. default = clone all comps.
		"""			
		if comps is None:
			comps = root.findChildren(tags=('TDAbleton',))
		sourceComps = self.SourceComp.findChildren(tags=('TDAbleton',))
		for source in sourceComps:
			source.par.Version = self.ownerComp.par.Version
		for comp in comps:
			allowCooking = comp.allowCooking
			if comp in sourceComps:
				continue
			if comp.par.clone.eval() in sourceComps:
				print('Update ' + comp.path)
				if hasattr(comp, 'Connected') and comp.Connected:
					comp.ext.TDAbletonCompBaseExt.removeAllListeners()
				comp.op('abletonBase').par.clone.val = \
									comp.op('abletonBase').par.clone.eval().path
				comp.allowCooking = False
				# fix version par name
				if comp.par['Tdabletonversion']:
					comp.par.Tdabletonversion.name = 'Version'
				comp.par.Version = self.ownerComp.par.Version
				parInfo = TDF.getParInfo(comp)
				fixInfo = {'compID': comp.id, 'allowCooking': allowCooking,
							'clone': comp.par.clone.expr}
				comp.par.enablecloningpulse.pulse()
				TDF.applyParInfo(comp, parInfo)

				# version inserts
				# force callback extension onto MIDI
				if comp.par.clone.eval().name == "abletonMIDI1":
					comp.par.extension2 = \
								"op('./CallbacksExt').module.CallbacksExt(me)"
					comp.par.promoteextension2 = True

				# postUpdate
				run("op(" + str(self.ownerComp.id) + ").ext.TDAbletonExt." \
									"fixUpdatedComp(" + str(fixInfo) + ")",
					delayFrames=1, delayRef=op.TDResources)
			elif comp.par.Tdabletoncomp.eval() == self.ownerComp:
				comp.SetError("Invalid clone source for update!")
		# for source in sourceComps:
		# 	source.par.enablecloning = False

	def fixUpdatedComp(self, info):
		comp = op(info['compID'])
		comp.allowCooking = info['allowCooking']
		# hack for par.enable staying linked to clone master when cloning off
		if comp.par.clone.eval() != comp \
				and comp.par.clone.mode == ParMode.EXPRESSION:
			#debug('clone', comp.path)
			comp.par.clone.val = ''
			comp.par.clone.mode = ParMode.CONSTANT
			run('args[0].par.enablecloningpulse.pulse()', comp, delayFrames=1,
					delayRef=op.TDResources)
			run(f"args[0].par.clone.mode = ParMode.EXPRESSION", comp, 
					delayFrames=2, delayRef=op.TDResources)
		#comp.par.reinitextensions.pulse()

	def NotifyAbletonComps(self, notificationType, infoDict=None):
		"""
		Call OnAbletonNotify on all Comps with tag TDAbleton

		onAbletonNotify has one argument which is info dictionary (with
			notification type added in: {'notificationType':<typestring>,...}
		:param type: the type of notification.
		:param info: extra info dictionary
		"""
		if infoDict is None:
			infoDict = {}
		infoDict['notificationType'] = notificationType
		for comp in self.TDAbletonComps:
			if hasattr(comp.ext, 'TDAbletonCompBaseExt') and comp.par.Connect:
				try:
					comp.ext.TDAbletonCompBaseExt.onAbletonNotify(infoDict)
				except:
					import traceback; traceback.print_exc()
					print('Error in onNotifyAbletonComp for ' + comp.path +
											'. Type:' + notificationType)

	def AddTDAMaxDevice(self, lomExpression, fileName,
												deviceName='', aParDict=''):
		"""
		Add a Max device from the TouchDesigner Remote Script folder

		:param lomExpression: expression for the chain to add device to
		:param fileName: the device file name
		:param deviceName: the device's new name ('' means no change)
		:param aParDict: changes to device parameters {<aPar name>: <value>...}
		:return: the device name after adding digits for duplicates...
		"""
		self.SendOSC('/tda/command', ['addDevice', fileName, lomExpression,
									  				deviceName, str(aParDict)])

	def RunRemoteCode(self, code):
		"""
		Run a chunk of python code on the associated Ableton Python Remote
		Script shell.
		:param code: a valid Python code string
		"""
		self.SendOSC('/shell/runCode', [code])

	def RequestRemoteData(self, code, callback, asRepr=False, includeID=False,
						  											id=None):
		"""
		Request data from Ableton Python Remote Script shell
		:param code: an evaluatable piece of Python code
		:param callback: the callback to call with return data
		:param asRepr: return data will be formatted by repr()
		:param includeID: callback will be called with a second argument: id
		:param id: id for returned data. if None, an int will be assigned
		:return: the id of this request
		"""
		if id is None:
			id = self.defaultRemoteDataID
			while id in self.remoteRequestDict:
				id += 1
				if id > 1000000:
					if len(self.remoteRequestDict) > 1000000:
						raise Exception("Maximum remote requests exceeded")
					id = 0
		self.remoteRequestDict[id] = (callback, asRepr, includeID)
		self.SendOSC('/shell/requestData', [code, int(asRepr), id])
		return id

	# endregion

	# region Callbacks

	def onReceiveOSCShellReply(self, dat, rowIndex, message, msgBytes,
							   timeStamp, address, args, peer):
		if address.endswith('/data'):
			if args[0] in self.remoteRequestDict:
				request = self.remoteRequestDict[args[0]]
				if request[2]:
					request[0](args[1], args[0])
				else:
					request[0](args[1])
		elif address.endswith('/reply'):
			self.Console.OnReply(args)

	def onReceiveOSCDebug(self, dat, rowIndex, message, msgBytes, timeStamp,
						  address, args, peer):
		if address.endswith('/ping'):
			print('TDAbleton Ping Response: \n'
				  '\tSent:   ', args[1], '\n\tReplied:', args[0],
				  '\n\tCurrent:', str(datetime.now()))
		elif address.endswith('/error'):
			relevantError = '\n\t'.join(args[0].splitlines())
			print('# TDA Remote Script Error #\n', relevantError,
				  '\n# End TDA Remote Script Error #')

	def onReceiveOSCInfo(self, dat, rowIndex, message, msgBytes, timeStamp,
						 address, args, peer):
		if address.endswith('/serverInit'):
			self._Connected.val = False
			if self.ownerComp.par.Nodeerrorondisconnect \
								and self.ownerComp.par.Connect:
				self.SetError("Disconnected from Ableton")
			if self.ownerComp.par.Connect.eval():
				self.connect()
		if address.endswith('serverDisconnect'):
			print('Ableton Live server closing')
			self.disconnect()
			if self.ownerComp.par.Connect.eval():
				self.connect()
		elif address.endswith('/songDump/start'):
			self.loadingSongInfo = ''
			try:
				self.retryConnectRepeater.kill()
			except:
				pass
			finally:
				self.retryConnectRepeater = None
		elif address.endswith('/songDump/chunk'):
			self.loadingSongInfo += args[0]
		elif address.endswith('/songDump/end'):
			self.onSongDumpComplete()
		elif address.endswith('/cuePointTimeChanged'):
			cuePointInfo = self.SongInfoByPtr['cuePoints'][args[0]]
			debug(cuePointInfo)
			cuePointInfo['time'] = args[1]
			self.SongInfo['cuePoints'][str(cuePointInfo['index'])] = \
														deepcopy(cuePointInfo)
			self.NotifyAbletonComps('cuePointTimeChanged', cuePointInfo)
		elif address.endswith('/clipSlotChanged'):
			clipSlotPtr = int(args[0])
			clipSlotInfo = self.ObjectInfoByPtr[clipSlotPtr]['byPtr']
			if len(args) > 1:
				clipName = args[1]
				clipFilePath = args[2]
				clipColor = args[3]
				clipSlotInfo['clip']['name'] = clipName
				clipSlotInfo['clip']['filePath'] = clipFilePath
				clipSlotInfo['clip']['color'] = clipColor
				clipSlotInfo['clip']['ptr'] = clipSlotPtr
			else:
				clipSlotInfo['clip'] = {}
			self.NotifyAbletonComps('clipSlotChanged', clipSlotInfo)
		elif address.endswith('/aClipChanged'):
			#debug('aClipChanged')
			aClipPtr = int(args[0])
			aClipInfo = self.ObjectInfoByPtr[aClipPtr]['byPtr']
			clipName = args[1]
			clipFilePath = args[2]
			clipColor = args[3]
			startTime = args[4]
			endTime = args[5]
			aClipInfo['name'] = clipName
			aClipInfo['filePath'] = clipFilePath
			aClipInfo['color'] = clipColor
			aClipInfo['startTime'] = startTime
			aClipInfo['endTime'] = endTime
			aClipInfo['ptr'] = aClipPtr
			self.NotifyAbletonComps('aClipChanged', aClipInfo)
		elif address.endswith('/updateArrangementClips'):
			#debug('updateAClips')
			chainInfo = self.ObjectInfoByPtr[args.pop(0)]['byPtr']
			aClips = chainInfo['aClips']
			aClips.clear()
			index = 0
			while(args):
				type = args.pop(0)
				if type == '#':
					aClipInfo = {}
					aClipInfo['name'] = args[0]
					aClipInfo['index'] = index
					aClipInfo['expression'] = f'arrangement_clip[{index}]'
					aClipInfo['startTime'] = args[1]
					aClipInfo['endTime'] = args[2]
					aClipInfo['startMarker'] = args[3]
					aClipInfo['endMarker'] = args[4]
					aClipInfo['loopStart'] = args[5]
					aClipInfo['loopEnd'] = args[6]
					aClipInfo['filePath'] = args[7]
					aClipInfo['color'] = args[8]
					aClipInfo['ptr'] = args[9]
					args = args[10:]
					aClipInfo['parentInfo'] = chainInfo
					self.ObjectInfoByPtr[int(aClipInfo['ptr'])] = {
							'byName': aClipInfo, 'byPtr': aClipInfo}	
					aClips.append(aClipInfo)
				else:
					args.pop(0)
			chainInfo['aClips'] = aClips
			self.NotifyAbletonComps('arrangementClipsChanged', chainInfo)
		elif address.endswith('/nameChanged'):
			ptr = int(args[0])
			name = args[1]
			objectInfoByPtr = self.ObjectInfoByPtr[ptr]['byPtr']
			objectInfoByName = self.ObjectInfoByPtr[ptr]['byName']
			oldName = objectInfoByPtr['name']


			# check for duplicate names before and after change
			parentDictKey = self.expressionToKey(objectInfoByPtr['expression'])
			if parentDictKey in ['tracks', 'cuePoints']:
				parentInfoByPtr = self.SongInfoByPtr
				parentInfoByName = self.SongInfo
				if parentDictKey == 'cuePoints':
					if oldName == name:
						return
					oldName = objectInfoByPtr['index']
			else:
				parentInfo = self.ObjectInfoByPtr[
										objectInfoByPtr['parentInfo']['ptr']]
				parentInfoByPtr = parentInfo['byPtr']
				parentInfoByName = parentInfo['byName']
			if oldName == name:
				return				
			duplicateNames = []
			for info in parentInfoByPtr[parentDictKey].values():
				if info['name'] == name:
					duplicateNames.append(info)
				if info['name'] == oldName and info['ptr'] != ptr:
					# there is a duplicate of this name
					# we need full song info, request and abort name change...
					self.SendOSC('/tda/command', ['songDump'])
					return
			if duplicateNames:
				duplicateNames.append(objectInfoByPtr)
				self.onDuplicateNames([duplicateNames])

			# fix entries
			objectInfoByPtr['name'] = name
			objectInfoByName['name'] = name
			if parentDictKey == 'cuePoints':
				parentInfoByName[parentDictKey][oldName] = objectInfoByName
			else:
				parentInfoByName[parentDictKey][name] = objectInfoByName
			del parentInfoByName[parentDictKey][oldName]
			self.NotifyAbletonComps('lomNameChanged', objectInfoByName)

		elif address.endswith('/debugging'):
			self.ownerComp.par.Logtdadebugmsgs = args[0]


	def SongPP(self, object):
		pprint.pprint(object, indent=4, sort_dicts=False, depth=2)

	def onParPulse(self, par):
		if par.name == 'Openconsole':
			self.Console.Open()
		elif par.name == 'Ping':
			self.SendOSC('/tda/ping', [str(datetime.now())])
		elif par.name == 'Updateabletoncomps':
			connected = self.ownerComp.par.Connect.eval()
			self.ownerComp.par.Connect = False
			try:
				self.UpdateAbletonComps()
			except:
				raise
			finally:
				if connected:
					self.ownerComp.par.Connect = True
			def onUpdateDialog(info):
				if info['button'] == 'Wiki':
					ui.viewFile('https://docs.derivative.ca/index.php?'
								'title=TDAbleton#Upgrading_TDAbleton')
			op.TDResources.op('popDialog').OpenDefault(
				'If you have saved any TDAbleton Max4Live devices locally in '
				'your Live Set, you may have to update them in Live. See Wiki '
				'for instructions.',
				"TDA Components Updated",
				buttons=['OK', 'Wiki'],
				callback=onUpdateDialog,
				textEntry=False,
				details=connected
				)
		# elif par.name in ('Helppage', 'Help'):
		# 	ui.viewFile('https://docs.derivative.ca/index.php?title=TDAbleton')
		elif par.name == 'Matchtimeline':
			self.MatchTimeline()
		elif par.name == 'Install':
			self.InstallTDAbleton()
		elif par.name == 'Opendemofolder':
			demoFolder = os.path.join(self.TDASamplesFolder, 'Demo Project')
			debug(demoFolder)
			ui.viewFile(demoFolder)
		elif par.name == 'Start':
			self.AutoStartLiveSet()
		elif par.name == 'Openlogfile':
			filepath = self.ownerComp.par.Abletonlogfile.eval()
			if app.osName == 'Windows':
				os.startfile(filepath)
			else:
				subprocess.call(('open', filepath))

	def onParValueChange(self, par, val, prev):
		if par.name == 'Version':
			for comp in self.ownerComp.op('comps').findChildren(
														tags=('TDAbleton',)):
				comp.par.Version = val
			self.NotifyAbletonComps("reinit")
			self.__init__(self.ownerComp)
		elif par.name == 'Abletonliveinstall':
			if menuOb:=self.AbletonLiveBuilds:
				testBuild = menuOb.menuNames[0]
				try:
					majorVersion = int(testBuild[5:].split('.')[0])
				except:
					majorVersion = 11
				if majorVersion < 11:
					def onVersionDialog(info):
						if info['button'] == 'Wiki':
							ui.viewFile('https://docs.derivative.ca/index.php?'
									'title=TDAbleton#'\
									'Install_the_latest_TDAbleton_system')
					op.TDResources.op('popDialog').OpenDefault(
						'For Ableton Live versions below 11, use TDAbleton/'
						'Live 9 & 10 in the palette.',
						"Requires Live 11+",
						buttons=['OK', 'Wiki'],
						callback=onVersionDialog,
						textEntry=False
						)
					self.ownerComp.par.Abletonliveinstall = \
														self.lastSelectedInstall
					return
			self.lastSelectedInstall = \
					self.ownerComp.par.Abletonliveinstall.eval()				
			self.ownerComp.cook(force=True)
		elif par.name == 'Logtdadbgmsgs':
			self.setDebugging()
		elif par.name == 'Connect':
			if par.eval():
				self.connect()
			else:
				self.disconnect()
		elif par.name in ['Inport', 'Abletonaddress', 'Abletonport', 'Maxport',
						  'Listenonly']:
			if not (par.name == 'Inport'
					and (self.TDABroadcast.op('out1')[0] is None
							or self.ownerComp.par.Abletonbroadcast ==
									self.TDABroadcast.op('out1')[0].eval())):
				# we skip the setup if broadcast has changed and we're waiting
				# for Live to update
				self.setupConnection()
		elif par.name == 'Abletonbroadcast':
			self.TDABroadcast.par.Valuesend = par.eval()
		elif par.name == 'Nodeerrorondisconnect':
			if par.eval():
				if not self.Connected and self.ownerComp.par.Connect:
					self.SetError("Disconnected from Ableton")
			else:
				if not self.Connected:
					self.ownerComp.clearScriptErrors(
											error='TDAbleton: Disconnected*')


	def onBroadcastChanged(self, val):
		broadcast = val
		if broadcast:
			self.ownerComp.par.Inport.expr = \
				'me.TDABroadcastPort.par.Valuesend'
		else:
			self.ownerComp.par.Inport.mode = ParMode.CONSTANT
		self.ownerComp.par.Inport.readOnly = broadcast
		if hasattr(self.ownerComp.parent, 'TDAbletonPackage'):
			self.ownerComp.parent.TDAbletonPackage.par.Inport.readOnly \
				= broadcast
		self.ownerComp.par.Abletonbroadcast = val
		self.setupConnection()

	# endregion

	# region Song Dump Processing

	def onSongDumpComplete(self):
		songInfo = TDJ.textToJSON(self.loadingSongInfo)
		liveTDAVersion = songInfo.get('tdaVersion', 'legacy')
		thisTDAVersion = self.ownerComp.par.Version.eval()
		if not self.Connected and liveTDAVersion != thisTDAVersion:
			def onVersionWarning(info):
				if info['button'] == 'Wiki':
					ui.viewFile('https://docs.derivative.ca/index.php?'
								'title=TDAbleton#Upgrading_TDAbleton')
			op.TDResources.op('popDialog').OpenDefault(
					'Live TDA Version: ' + liveTDAVersion + '\n'
					'This TDA Version: ' + thisTDAVersion + '\n'
					'Some features may not work.\n\n'
					'To fix this, upgrade to the latest TDAbleton.',
										"TDAbleton Version Mismatch",
					buttons=['OK','Wiki'], 
					callback=onVersionWarning)
		self.ObjectInfoByPtr.clear()
		self.SongInfoByPtr.clear()
		self.SongInfoByPtr.update(songInfo)
		# create SongInfo dictionary
		self.SongInfo.clear()
		self.SongInfo.update(songInfo)
		self.SongInfo['tracks'] = dict()
		self.SongInfo['cuePoints'] = dict()
		self.SongInfo['scenes'] = []
		duplicateNames = [] # store all duplicate names for error report
		# create SongInfo (by name)
		# also convert ptr strings to ints in SongInfoByPtr
		# dive down into tracks/chains
		trackNames = {}
		for trackInfo in list(self.SongInfoByPtr['tracks'].values()):
			trackInfo['parentInfo'] = None
			# store for duplicate names
			name = trackInfo['name']
			if name in trackNames:
				trackNames[name].append(trackInfo)
			else:
				trackNames[name] = [trackInfo]

			self.convertSongDumpChain(trackInfo,
									self.SongInfo['tracks'], duplicateNames)
			for clipSlotInfo in trackInfo['clipSlots']:
				clipSlotInfo['parentInfo'] = trackInfo
				self.ObjectInfoByPtr[int(clipSlotInfo['ptr'])] = {
						'byName': clipSlotInfo, 'byPtr': clipSlotInfo}
				if clipSlotInfo['clip']:
					clip = clipSlotInfo['clip']
					self.ObjectInfoByPtr[int(clip['ptr'])] = {
									'byName': clip,
									'byPtr': clip}
			if 'aClips' in trackInfo:
				for aClipInfo in trackInfo['aClips']:
					aClipInfo['parentInfo'] = trackInfo
					self.ObjectInfoByPtr[int(aClipInfo['ptr'])] = {
							'byName': aClipInfo, 'byPtr': aClipInfo}	
		for cuePointInfo in list(self.SongInfoByPtr['cuePoints'].values()):
			byName = self.SongInfo['cuePoints'][str(cuePointInfo['index'])] = \
													dict(cuePointInfo)
			self.ObjectInfoByPtr[int(cuePointInfo['ptr'])] = {
					'byName': byName, 'byPtr': cuePointInfo}
		for sceneInfo in list(self.SongInfoByPtr['scenes'].values()):
			self.SongInfo['scenes'].append(dict(sceneInfo))
			byName = self.SongInfo['scenes'][-1]
			self.ObjectInfoByPtr[int(sceneInfo['ptr'])] = {
					'byName': byName, 'byPtr': sceneInfo}

		# notify other ableton comps
		if self.Connected:
			self.NotifyAbletonComps('songInfo')
		else:
			if self.SongInfo['name'] is not None:
				print(self.ownerComp.path,
					  ': Connected to Ableton Live. Song: "' +
												self.SongInfo['name'] + '".')
			else:
				print(self.ownerComp.path, ': Connected to Ableton Live.')

			self.connectingDialog.par.Close.pulse()

			self._Connected.val = True
			self.SetError()

			self.NotifyAbletonComps('connected')
		# check for master
		if self.SongInfo['name'] is None:
			self.onNoMaster()
			self._Connected.val = False
			return
		# duplicate name check
		self.checkNames(trackNames, duplicateNames)
		self.onDuplicateNames(duplicateNames)
		# setup ops

		# self.TDABroadcast.par.Device = self.TDAMasterDevice['ptr']
		# self.TDABroadcastPort.par.Device = self.TDAMasterDevice['ptr']
		# self.TDAAbletonSendTime.par.Device = self.TDAMasterDevice['ptr']
		run('args[0].setMasters()', self, delayFrames=1, 
				delayRef=op.TDResources)

		# set Ableton version in utilities if possible
		try:
			version = self.SongInfo['filepath']
			for folder in range(4):
				version = os.path.dirname(version)
			version = os.path.split(version)[1]
			if version in self.ownerComp.par.Abletonliveinstall.menuNames:
				self.ownerComp.par.Abletonliveinstall = version
		except:
			pass

	def setMasters(self):
		self.TDABroadcast.par.Device = self.TDAMasterDevice['ptr']
		self.TDABroadcastPort.par.Device = self.TDAMasterDevice['ptr']
		self.TDAAbletonSendTime.par.Device = self.TDAMasterDevice['ptr']

	def convertSongDumpChain(self, chainDictByPtr, chainDictsByName,
							 								duplicateNames):
		"""
		Fixall ptrs and create SongInfo by name for a chain. Tracks count as
		chains for this purpose
		"""
		chainDictByName = chainDictsByName[chainDictByPtr['name']] = \
			dict(chainDictByPtr)
		self.ObjectInfoByPtr[int(chainDictByPtr['ptr'])] = {
					'byName': chainDictByName, 'byPtr': chainDictByPtr}
		chainDictByName['devices'] = dict()
		deviceNames = {}
		for deviceDictByPtr in list(chainDictByPtr['devices'].values()):
			deviceDictByPtr['parentInfo'] = chainDictByName
			# store for duplicate names
			name = deviceDictByPtr['name']
			if name in deviceNames:
				if deviceDictByPtr['class'] == 'PluginDevice':
					index = 2
					while f'{name} <{index}>)' in deviceNames:
						index += 1
					name = f'{name} <{index}>)'
					deviceDictByPtr['name'] = name
					deviceNames[name] = [deviceDictByPtr]
				else:
					deviceNames[name].append(deviceDictByPtr)
			else:
				deviceNames[name] = [deviceDictByPtr]

			deviceDictByName = \
					chainDictByName['devices'][deviceDictByPtr['name']] = \
							dict(deviceDictByPtr)
			self.ObjectInfoByPtr[int(deviceDictByPtr['ptr'])] = {
					 'byName': deviceDictByName, 'byPtr': deviceDictByPtr}
			deviceDictByName['aPars'] = dict()
			for aParDictByPtr in list(deviceDictByPtr['aPars'].values()):
				aParDictByPtr['parentInfo'] = deviceDictByName
				aParDictByName = \
						deviceDictByName['aPars'][aParDictByPtr['name']] = \
								dict(aParDictByPtr)
				self.ObjectInfoByPtr[int(aParDictByPtr['ptr'])] = {
						'byName': aParDictByName,
						'byPtr': aParDictByPtr}
			deviceDictByName['chains'] = dict()
			chainNames = {}
			for deviceChainDictByPtr in \
									list(deviceDictByPtr['chains'].values()):
				deviceChainDictByPtr['parentInfo'] = deviceDictByName
				# store for duplicate names
				name = deviceChainDictByPtr['name']
				if name in chainNames:
					chainNames[name].append(deviceChainDictByPtr)
				else:
					chainNames[name] = [deviceChainDictByPtr]
				self.convertSongDumpChain(deviceChainDictByPtr,
								  deviceDictByName['chains'], duplicateNames)
			self.checkNames(chainNames, duplicateNames)
		self.checkNames(deviceNames, duplicateNames)
		return chainDictByName

	@staticmethod
	def expressionToKey(expression):
		"""
		return the key name in parent object for the dictionary holding the
			object with the provided expression.
		e.g. if an object's expression is 'cue_points[3]', this returns
			'cuePoints'
		"""
		if expression.startswith('parameters'):
			return 'aPars'
		word = expression.split('[')[0]
		return word.split('_')[0] + ''.join(
							x.capitalize() or '_' for x in word.split('_')[1:])

	@staticmethod
	def checkNames(nameDict, duplicateNames):
		for name, info in nameDict.items():
			if len(info) > 1:
				duplicateNames.insert(0, info)

	def onNoMaster(self):
		self.SetError('No TDA Master device found in Live. See dialog.')
		dialogReport = "A TDA Master device is required on Live's master " \
					   'track.\n\nEnter a song name and press' \
					   ' "Create" to add device. "Cancel" will cause ' \
					   'TDAbleton to abort connection.'
		op.TDResources.op('popDialog').Open(dialogReport,
										"No TDA Master",
										textEntry='TD Song',
										buttons=['Create', "Cancel"],
										callback=self.onNoMasterPopup)

	def onNoMasterPopup(self, info):
		if info['button'] == "Cancel":
			return
		if not info['enteredText'].strip():
			info['enteredText'] = 'TD Song'
		self.AddTDAMaxDevice('SONG.master_track', 'TDA_Master.amxd',
										  				 info['enteredText'])

	def onDuplicateNames(self, duplicates):
		if not self.ownerComp.par.Duplicatenames.menuIndex:
			return
		dialogReport = ''
		textportReport = '\nTDAbleton Error - '
		for dupList in duplicates:
			name = dupList[0]['name']
			msg = '\nDuplicate: "' + name + '"'
			dialogReport += msg
			# textport only details
			textportReport += msg
			textportReport += ' (' + ', '.join(
							[d['expression'] for d in dupList]) + ')'
			path = ''
			info = dupList[0]
			while info['parentInfo']:
				info = info['parentInfo']
				path = '\n    ' + info['name'] + \
						 ' (' + info['expression'] + ')' + path
			if path:
				textportReport += ' in path:' + path
		textportReport += '\n(For more options, see parameter ' \
						  		'"Duplicate Names" on tdAbleton Component)'
		if dialogReport:
			heading = 'Duplicate names will cause problems in TDAbleton. '\
					  'Please make sure the following names are unique. ' \
					  '\n(See textport for details)\n'
			dialogReport = heading + dialogReport
			print(textportReport + '\n')
			if self.ownerComp.par.Duplicatenames.menuIndex == 2:
				op.TDResources.op('popDialog').Open(dialogReport,
											"Duplicate Names Found",
											textEntry=False,
									   		buttons=['Ignore', 'Autofix'],
											callback=self.onDuplicateNamesPopup,
											details=duplicates)
			elif self.ownerComp.par.Duplicatenames.menuIndex == 3:
				self.autoFixDuplicateNames(duplicates)

	def onDuplicateNamesPopup(self, info):
		if info['button'] == 'Autofix':
			self.autoFixDuplicateNames(info['details'])

	def autoFixDuplicateNames(self, duplicates):
		autofixCommand = ''
		for dupList in duplicates:
			info = dupList[0]
			path = ''
			pathInfo = info
			name = pathInfo['name']
			while pathInfo['parentInfo']:
				pathInfo = pathInfo['parentInfo']
				path = pathInfo['expression'] + '.' + path
			path = 'SONG.' + path
			appendNum = 2
			listExpression = info['expression'][:info['expression'].index('[')]
			if listExpression == 'parameters':
				listExpression = 'aPars'
			parentInfo = info['parentInfo']
			for info in dupList[1:]:
				newName = name + '-' + str(appendNum)
				if parentInfo:
					while newName in parentInfo[listExpression].keys():
						appendNum += 1
						newName = name + '-' + str(appendNum)
					parentInfo[listExpression][newName] = info
				changePath = path + info['expression'] + '.name = "' + \
							 newName +'"'
				autofixCommand += changePath + '\n'
		self.RunRemoteCode(autofixCommand)
		print('Duplicate name autofix commands sent:\n' + autofixCommand)
	# endregion

	# region Auto Install

	@property
	def AbletonLiveInstalls(self):
		installs = ['Custom']
		installNames = ['0']
		if app.osName == 'Windows':
			AbletonPath = os.path.join(os.getenv('PROGRAMDATA'), 'Ableton')
			try:
				for f in os.scandir(AbletonPath):
					if f.is_dir() and f.name.startswith('Live'):
						installs.insert(1, f.name)
						installNames.insert(1, f.name)
			except:
				pass
		else:
			try:
				for f in os.scandir('/Applications'):
					if f.is_dir() and f.name.startswith('Ableton'):
							installs.insert(1, f.name)
							installNames.insert(1, f.name)
			except:
				pass
		return TDF.parMenu(installNames, installs)

	@property
	def DefaultRemoteScriptFolder(self):
		install = self.ownerComp.par.Abletonliveinstall.eval()
		if not install or install == '0':
			return "No Ableton Install Selected"
		if app.osName == 'Windows':
			installPath = os.path.join(os.getenv('PROGRAMDATA'),
									   'Ableton',
									   install,
									   'Resources',
									   'MIDI Remote Scripts'
									   )
		else:
			installPath = os.path.join('/Applications',
									   install,
									   'Contents',
									   'App-Resources',
									   'MIDI Remote Scripts'
									   )

		return installPath

	@property
	def AbletonLiveBuilds(self):
		#readOnly = self.ownerComp.par.Abletonlivebuild.readOnly
		install = self.ownerComp.par.Abletonliveinstall.eval()
		if not install or install == '0':
			self.ownerComp.par.Abletonlivebuild.readOnly = True
			run('args[0].menuIndex = args[1]', 
						self.ownerComp.par.Abletonlivebuild,
						0, delayFrames=1, delayRef=op.TDResources)
			return TDF.parMenu(['No Ableton Install Selected', 
						self.ownerComp.par.Abletonlivebuild.eval()])
		try:
			prefPath = self.PrefPath
			if app.osName == 'Windows':
				installPrefix = install[:install.find(' ', 5)]
			else:
				installPrefix = install[install.find('Live'):install.find(' ',
													install.find('Live') + 5)]
			dirs = [f.name for f in os.scandir(prefPath)
					if f.name.startswith(installPrefix) and f.name[5].isdigit()]
		except:
			self.ownerComp.par.Abletonlivebuild.readOnly = True
			run('args[0].menuIndex = args[1]', 
						self.ownerComp.par.Abletonlivebuild,
						0, delayFrames=1, delayRef=op.TDResources)
			return TDF.parMenu([
					"No Build Info Found.  Try opening then closing Live."])
		try:
			dirs.sort(key=lambda x: version.LooseVersion(x[5:]) 
								if 'b' not in x else version.LooseVersion('0'))
		except:
			if self.ownerComp.par.Abletonlivebuild.readOnly:
				op.TDResources.op('popDialog').OpenDefault(
					"Unable to determine which build, probably due to Live "
					"Betas on the system. Please be sure to set the Ableton "
					"Live Build parameter.", "Problem Selecting Build")
				run('args[0].menuIndex = args[1]', 
						self.ownerComp.par.Abletonlivebuild,
						100, delayFrames=1, delayRef=op.TDResources)
				self.ownerComp.par.Abletonlivebuild.readOnly = True
		else:
			self.ownerComp.par.Abletonlivebuild.readOnly = \
											not any(['b' in x for x in dirs])
			run('args[0].menuIndex = args[1]', 
						self.ownerComp.par.Abletonlivebuild,
						len(dirs)-1, delayFrames=1, delayRef=op.TDResources)
			if self.ownerComp.parent().par['Abletonlivebuild'] is not None:
				self.ownerComp.parent().par['Abletonlivebuild'].readOnly = \
					self.ownerComp.par.Abletonlivebuild.readOnly
		return TDF.parMenu(dirs)			

	@property
	def DefaultPreferencesFolder(self):
		install = self.ownerComp.par.Abletonliveinstall.eval()
		if not install or install == '0':
			return "No Ableton Install Selected"
		buildDir = self.ownerComp.par.Abletonlivebuild.eval()
		if app.osName == 'Windows':
			return os.path.join(self.PrefPath, buildDir, 'Preferences')
		else:
			return os.path.join(self.PrefPath, buildDir)

	@property
	def LogFile(self):
		prefFolder = self.DefaultPreferencesFolder
		if prefFolder.startswith(("No Ableton Install", 
												"No Preferences Folder Found")):
			return prefFolder
		else:
			return os.path.join(prefFolder, 'Log.txt')

	@property
	def PrefPath(self):
		if app.osName == 'Windows':
			prefPath = os.path.join(os.getenv('APPDATA'),
									'Ableton')
		else:
			userFolder = os.path.expanduser('~')
			prefPath = os.path.join(userFolder, 'Library', 'Preferences',
									'Ableton')
		return prefPath

	def InstallTDAbleton(self):
		remoteScriptFolder = self.ownerComp.par.Remotescriptfolder.eval()
		prefFolder = self.ownerComp.par.Preferencesfolder.eval()
		tdRemoteScriptFolder = os.path.join(remoteScriptFolder, 'TouchDesigner')
		def onInstallOverwrite(info):
			if info['button'] == 'Cancel':
				msg = 'TDAbleton install cancelled'
				ui.status = msg
				print(msg)
			else:
				self.doInstallTDAbleton()

		if prefFolder.startswith('No Preferences Folder Found'):
			op.TDResources.op('popDialog').OpenDefault(
				"No Ableton Live preferences folder found. If you have never "
				"run this Live installation, open it then close it again to "
				"create a Preferences folder.", "No Preferences Folder")
			return

		op.TDResources.op('popDialog').Open(
			"Make sure Ableton Live is closed before installing.\n\n"
			"This will copy the latest TDAbleton MIDI Remote Script into Live "
			"and set up the TouchDesigner User Folder in Live's browser.",
			"Install TDAbleton",
			textEntry=False,
			buttons=['Confirm', 'Cancel'],
			enterButton=1,
			escButton=2,
			details=tdRemoteScriptFolder,
			callback=onInstallOverwrite)

	def doInstallTDAbleton(self):
		remoteScriptFolder = self.ownerComp.par.Remotescriptfolder.eval()
		tdRemoteScriptFolder = os.path.join(remoteScriptFolder, 'TouchDesigner')
		tdAbletonFolder = os.path.join(self.TDASamplesFolder, 'TouchDesigner')
		def onDeleteDialog(info):
			if info['button'] == 'Folder':
				ui.viewFile(tdRemoteScriptFolder, showInFolder=True)		
		if os.path.exists(tdRemoteScriptFolder):
			try:
				shutil.rmtree(tdRemoteScriptFolder)
			except:
				op.TDResources.op('popDialog').OpenDefault(
					f'Failed to remove MIDI Remote Script folder. '\
					'Run TouchDesigner as admin or delete '\
					'TouchDesigner Remote Script folder by hand, '\
					'then try again.',
					"Error Deleting Old Version", buttons=['OK', 'Folder'],	
					textEntry=False, callback=onDeleteDialog)	
				return
		shutil.copytree(tdAbletonFolder, tdRemoteScriptFolder)
		print('\nTouchDesigner folder copied to ' + remoteScriptFolder + '.')

		prefFolder = self.ownerComp.par.Preferencesfolder.eval()
		configFile = os.path.join(prefFolder, 'Library.cfg')
		et = ElementTree.parse(configFile)
		userFolder = et.getroot().findall(
								'./ContentLibrary/UserFolderInfoList')[0]
		remoteScriptFolder = self.ownerComp.par.Remotescriptfolder.eval()
		remoteScriptFolderXML = \
						os.path.join(remoteScriptFolder, 'TouchDesigner', '')
		lastId = 0
		tdFolder = None
		for folder in userFolder.findall('UserFolderInfo'):
			if folder.get('DisplayName') == 'TouchDesigner':
				tdFolder = folder
				break
			id = int(folder.get('Id'))
			if id > lastId:
				lastId = id
		if tdFolder is None:
			ElementTree.SubElement(userFolder, 'UserFolderInfo',
								   {'Path':remoteScriptFolderXML,
									'Id':str(lastId+1),
									'DisplayName': 'TouchDesigner'})
		else:
			tdFolder.set('Path', remoteScriptFolderXML)
		et.write(configFile, encoding='utf-8', xml_declaration=True)
		print('TouchDesigner entry added to Ableton Live User Folders.')

		msg = 'TDAbleton installed successfully.'
		ui.status = msg
		print(msg)
		op.TDResources.op('popDialog').OpenDefault(
			"TDAbleton installed successfully. \n\nFurther Instructions:\n"
			"If you haven't already, add "
			'the TouchDesigner Control Script in Live\'s "Link '
			'MIDI" section in Preferences. To update components deployed in '
			'this project, '
			'click the Update Ableton Comps parameter.',
			"TDAbleton Install Successful",
			textEntry=False,
			)

	def BinaryXML(self):
		prefFolder = self.ownerComp.par.Preferencesfolder.eval()
		configFile = os.path.join(prefFolder, 'Preferences.cfg')
		f = open(configFile, 'r')
		#et = ElementTree.parse(configFile)
		return f

	@property
	def TDASamplesFolder(self):
		folder = app.samplesFolder
		if self.ownerComp.par.Version.eval()[-1] == 'b':
			return os.path.join(folder, 'TDAbletonBeta', 'Live 11')
		else:
			return os.path.join(folder, 'TDAbleton', 'Live 11')

	def AutoStartLiveSet(self):
		install = self.ownerComp.par.Abletonliveinstall.eval()
		file = os.path.abspath(self.ownerComp.par.Autostartliveset.eval())
		error = None
		if not install or install == '0':
			error = 'No install specified'
		elif not (liveSet:=self.ownerComp.par.Autostartliveset.eval()):
			error = 'No live set specified'
		elif not os.path.exists(file):
			error = 'Auto-Start Live Set does not exist'
		if error:
			op.TDResources.op('popDialog').OpenDefault(error, 
										'Auto-Start Live Set Error')
			return
		if app.osName == 'Windows':
			appPath = os.path.join(os.getenv('PROGRAMDATA'),
										'Ableton',
										install,
										'Program',
										'Ableton ' + install)
		else:
			appPath = os.path.join('/Applications',
									   install,
									   'Contents',
									   'MacOS',
									   'Live')
		command = [appPath, file]
		result = subprocess.Popen(command)
	# endregion

	pass
